package com.sleepycat.je.tree;

import java.io.PrintStream;

import com.sleepycat.je.Cursor;
import com.sleepycat.je.DatabaseEntry;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.DbInternal;
import com.sleepycat.je.LockMode;
import com.sleepycat.je.OperationStatus;
import com.sleepycat.je.dbi.CursorImpl;
import com.sleepycat.je.dbi.DatabaseImpl;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.dbi.GetMode;
import com.sleepycat.je.tree.Tree.SearchType;
import com.sleepycat.je.txn.LockType;
import com.sleepycat.je.txn.Locker;
import com.sleepycat.je.txn.ThreadLocker;

public privileged aspect TreeVisitorAspect {

	
	private TreeVisitor Tree.getTreeVisitor() {
		if (EnvironmentImpl.getThreadLocalReferenceCount() > 0) {
			return (TreeVisitor) treeVisitorTL.get();
		} else {
			return null;
		}
	}

	public void Tree.setTreeVisitor(TreeVisitor tSA) {
		treeVisitorTL.set(tSA);
	}

	private ThreadLocal Tree.treeVisitorTL = new ThreadLocal();

	after(IN nextIN, Tree tree):call(void Tree.hook_foundLeafBIN(IN)) && args(nextIN) && target(tree) {
		TreeVisitor treeVisitor = tree.getTreeVisitor();
		if (treeVisitor != null) {
			nextIN.accept(treeVisitor);
		}
	}

	after(Tree tree) returning(Node child):
		call(Node IN.fetchTarget(int)) && 
		withincode(IN Tree.searchSubTree(IN, byte[], SearchType, long, BINBoundary, boolean)) && 
		this(tree) 
	{
		TreeVisitor treeVisitor = tree.getTreeVisitor();
		if (treeVisitor != null) {
			((IN) child).accept(treeVisitor);
		}
	}

	before(IN parent, Tree tree):
		call(void Tree.hook_beforeLoopChildren(IN)) && args(parent) && target(tree) 
	{
		TreeVisitor treeVisitor = tree.getTreeVisitor();
		if (treeVisitor != null) {
			parent.accept(treeVisitor);
		}

	}

	public void CursorImpl.setTreeVisitor(TreeVisitor tSA) {
		treeVisitorTL.set(tSA);
	}

	private ThreadLocal CursorImpl.treeVisitorTL = new ThreadLocal();

	private TreeVisitor CursorImpl.getTreeVisitor() {
		if (EnvironmentImpl.getThreadLocalReferenceCount() > 0) {
			return (TreeVisitor) treeVisitorTL.get();
		} else {
			return null;
		}
	}

	before(CursorImpl cursorImpl):
		execution(void CursorImpl.incrementLNCount()) && target(cursorImpl){

		TreeVisitor treeVisitor = cursorImpl.getTreeVisitor();
		if (treeVisitor != null) {
			treeVisitor.incrementLNCount();
		}
	}

	after(CursorImpl cursorImpl) throws DatabaseException:
		execution(void CursorImpl.hook_walker()) && target(cursorImpl){

		TreeVisitor treeStatsAccumulator = cursorImpl.getTreeVisitor();
		if (treeStatsAccumulator != null) {
			// Lck latchBIN();
			try {


				DIN duplicateRoot = (DIN) cursorImpl.bin.fetchTarget(cursorImpl.index);
				// Lck duplicateRoot.latch();
				try {
					DupCountLN dcl = duplicateRoot.getDupCountLN();
					if (dcl != null) {
						dcl.accept(treeStatsAccumulator);
					}
				} finally {
					// Lck duplicateRoot.releaseLatch();
				}
			} finally {
				// Lck releaseBIN();
			}
		}
	}

	before(Node n, CursorImpl cursorImpl):
		execution(void CursorImpl.hook_doesNotContainDuplicates(Node)) && args(n) && target(cursorImpl){
		/*
		 * Even if the entry is deleted, just leave our position here and
		 * return.
		 */
		TreeVisitor treeVisitor = cursorImpl.getTreeVisitor();
		if (treeVisitor != null) {
			if (n == null || ((LN) n).isDeleted()) {
				treeVisitor.incrementDeletedLNCount();
			} else {
				treeVisitor.incrementLNCount();
			}
		}

	}

	after(CursorImpl cursorImpl) returning (OperationStatus os):
		execution(OperationStatus CursorImpl.fetchCurrent(DatabaseEntry, DatabaseEntry, LockType, boolean))
		&& target(cursorImpl)
	{

		if (os == OperationStatus.KEYEMPTY) {
			TreeVisitor treeVisitor = cursorImpl
					.getTreeVisitor();
			if (treeVisitor != null) {
				treeVisitor.incrementDeletedLNCount();
			}
		}
	}
	
	/*
	 * DbStat support.
	 */
	void DBIN.accept(TreeVisitor acc) {
		acc.processDBIN(this, new Long(getNodeId()), getLevel());
	}
	
	/*
	 * DbStat support.
	 */
	void DIN.accept(TreeVisitor acc) {
		acc.processDIN(this, new Long(getNodeId()), getLevel());
	}
	
	/*
	 * DbStat support.
	 */
	public void DupCountLN.accept(TreeVisitor acc) {
		acc.processDupCountLN(this, new Long(getNodeId()));
	}
	
	/*
	 * DbStat support.
	 */
	void BIN.accept(TreeVisitor acc) {
		acc.processBIN(this, new Long(getNodeId()), getLevel());
	}
	
	/*
	 * DbStat support.
	 */
	void IN.accept(TreeVisitor acc) {
		acc.processIN(this, new Long(getNodeId()), getLevel());
	}
	
	
	/*
	 * @return true if no errors.
	 */
	private boolean DatabaseImpl.walkDatabaseTree(TreeVisitor visitor,
			PrintStream out, boolean verbose) throws DatabaseException {

		boolean ok = true;
		Locker locker = new ThreadLocker(envImpl);
		Cursor cursor = null;
		CursorImpl impl = null;
		try {
			EnvironmentImpl.incThreadLocalReferenceCount();
			cursor = DbInternal.newCursor(this, locker, null);
			impl = DbInternal.getCursorImpl(cursor);
			tree.setTreeVisitor(visitor);

			/*
			 * This will only be used on the first call for the position() call.
			 */
			impl.setTreeVisitor(visitor);
			DatabaseEntry foundData = new DatabaseEntry();
			DatabaseEntry key = new DatabaseEntry();
			OperationStatus status = DbInternal.position(cursor, key,
					foundData, LockMode.READ_UNCOMMITTED, true);
			while (status == OperationStatus.SUCCESS) {
				try {
					status = DbInternal.retrieveNext(cursor, key, foundData,
							LockMode.READ_UNCOMMITTED, GetMode.NEXT);
				} catch (DatabaseException DBE) {
					ok = false;
					if (DbInternal.advanceCursor(cursor, key, foundData)) {
						if (verbose) {
							out.println("Error encountered (continuing):");
							out.println(DBE);
							printErrorRecord(out, key, foundData);
						}
					} else {
						throw DBE;
					}
				}
			}
		} finally {
			if (impl != null) {
				impl.setTreeVisitor(null);
			}
			tree.setTreeVisitor(null);
			EnvironmentImpl.decThreadLocalReferenceCount();
			if (cursor != null) {
				cursor.close();
			}
		}

		return ok;
	}
	
	
	/*
	 * ThreadLocal.get() is not cheap so we want to minimize calls to it. We
	 * only use ThreadLocals for the TreeStatsAccumulator which are only called
	 * in limited circumstances. Use this reference count to indicate that a
	 * thread has set a TreeStatsAccumulator. When it's done, it decrements the
	 * counter. It's static so that we don't have to pass around the
	 * EnvironmentImpl.
	 */
	private static int EnvironmentImpl.threadLocalReferenceCount = 0;

	public static int EnvironmentImpl.getThreadLocalReferenceCount() {
		return threadLocalReferenceCount;
	}

	public static synchronized void EnvironmentImpl.incThreadLocalReferenceCount() {
		threadLocalReferenceCount++;
	}

	public static synchronized void EnvironmentImpl.decThreadLocalReferenceCount() {
		threadLocalReferenceCount--;
	}

}